import attrs
import typing as ty
from collections import OrderedDict
import logging
from copy import copy
from typing_extensions import Self
import attrs.converters
from fileformats.core import DataType
from fileformats.core.exceptions import FormatConversionError
from pydra.utils.typing import StateArray
from pydra.utils import get_fields
from pydra.compose.base import Task
from pydra.compose import workflow, python
from frametree.core.exceptions import (
    FrameTreeNameError,
    FrameTreeUsageError,
    FrameTreeDesignError,
    FrameTreePipelinesStackError,
    FrameTreeOutputNotProducedException,
    FrameTreeDataMatchError,
)
import frametree.core.frameset.base
import frametree.core.row
from frametree.core.axes import Axes
from frametree.core.utils import (
    path2varname,
    add_exc_note,
)
from frametree.core.serialize import (
    asdict,
    fromdict,
    pydra_asdict,
    pydra_fromdict,
    ClassResolver,
    ObjectListConverter,
)
from frametree.core.column import SinkColumn
from frametree.core.frameset.base import FrameSet


logger = logging.getLogger("frametree")


@attrs.define
class PipelineField:
    """Defines an input to a pipeline

    Parameters
    ----------
    name : str
        Name of the input and how it will be referred to in UI
    field : str, optional
        the name of the pydra input field to connect to, defaults to name
    datatype : type, optional
        the type of the items to be passed to the input, fileformats.generic.File by default
    """

    name: str
    field: str = attrs.field()
    datatype: type = attrs.field(
        default=None,
        converter=ClassResolver(
            DataType,
            allow_none=True,
            alternative_types=[frametree.core.row.DataRow],
        ),
    )

    @field.default
    def field_default(self) -> str:
        return self.name


logger = logging.getLogger("frametree")


@attrs.define
class Pipeline:
    """A thin wrapper around a Pydra workflow to link it to sources and sinks
    within a frameset

    Parameters
    ----------
    name : str
        the name of the pipeline, used to differentiate it from others
    row_frequency : Axes, optional
        The row_frequency of the pipeline, i.e. the row_frequency of the
        derivatvies within the frameset, e.g. per-session, per-subject, etc,
        by default None
    workflow : Workflow
        The pydra workflow that performs the actual analysis
    inputs : Sequence[ty.Union[str, ty.Tuple[str, type]]]
        List of column names (i.e. either data sources or sinks) to be
        connected to the inputs of the pipeline. If the pipelines requires
        the input to be in a datatype to the source, then it can be specified
        in a tuple (NAME, FORMAT)
    outputs : Sequence[ty.Union[str, ty.Tuple[str, type]]]
        List of sink names to be connected to the outputs of the pipeline
        If the input to be in a specific datatype, then it can be provided in
        a tuple (NAME, FORMAT)
    converter_args : dict[str, dict]
        keyword arguments passed on to the converter to control how the
        conversion is performed.
    frameset : FrameSet
        the frameset the pipeline has been applied to
    """

    name: str = attrs.field()
    row_frequency: Axes = attrs.field()
    task: Task = attrs.field()
    inputs: ty.List[PipelineField] = attrs.field(
        converter=ObjectListConverter(PipelineField)
    )
    outputs: ty.List[PipelineField] = attrs.field(
        converter=ObjectListConverter(PipelineField)
    )
    converter_args: ty.Dict[str, dict] = attrs.field(
        factory=dict, converter=attrs.converters.default_if_none(factory=dict)
    )
    frameset: frametree.core.frameset.base.FrameSet = attrs.field(
        metadata={"asdict": False}, default=None, eq=False, hash=False
    )

    def __attrs_post_init__(self) -> None:
        for field in self.inputs + self.outputs:
            if field.datatype is None:
                field.datatype = self.frameset[field.name].datatype

    @inputs.validator
    def inputs_validator(self, _: ty.Any, inputs: ty.List[PipelineField]) -> None:
        for inpt in inputs:
            if inpt.datatype is frametree.core.row.DataRow:  # special case
                continue
            if self.frameset:
                column = self.frameset[inpt.name]
                # Check that a converter can be found if required
                if (
                    inpt.datatype
                    and not issubclass(inpt.datatype, column.datatype)
                    and not issubclass(column.datatype, inpt.datatype)
                ):
                    try:
                        inpt.datatype.get_converter(column.datatype)
                    except FormatConversionError as e:
                        msg = (
                            f"required to in conversion of '{inpt.name}' input "
                            f"to '{self.name}' pipeline"
                        )
                        add_exc_note(e, msg)
                        raise
            elif inpt.datatype is None:
                raise ValueError(
                    f"Datatype must be explicitly set for {inpt.name} in unbound Pipeline"
                )
            field_names = [f.name for f in get_fields(self.task)]
            if inpt.field not in field_names:
                raise FrameTreeNameError(
                    inpt.field,
                    f"{inpt.field} is not in the input spec of '{self.name}' "
                    f"pipeline: " + "', '".join(field_names),
                )

    @outputs.validator
    def outputs_validator(self, _: ty.Any, outputs: ty.List[PipelineField]) -> None:
        for outpt in outputs:
            if self.frameset:
                column = self.frameset[outpt.name]
                if column.row_frequency != self.row_frequency:
                    raise FrameTreeUsageError(
                        f"Pipeline row_frequency ('{str(self.row_frequency)}') doesn't match "
                        f"that of '{outpt.name}' output ('{str(self.row_frequency)}')"
                    )
                # Check that a converter can be found if required
                if (
                    outpt.datatype
                    and not issubclass(outpt.datatype, column.datatype)
                    and not issubclass(column.datatype, outpt.datatype)
                ):
                    try:
                        column.datatype.get_converter(outpt.datatype)
                    except FormatConversionError as e:
                        msg = (
                            f"required to in conversion of '{outpt.name}' output "
                            f"from '{self.name}' pipeline"
                        )
                        add_exc_note(e, msg)
                        raise
            elif outpt.datatype is None:
                raise ValueError(
                    f"Datatype must be explicitly set for {outpt.name} in unbound Pipeline"
                )
            field_names = [f.name for f in get_fields(self.task.Outputs)]
            if outpt.field not in field_names:
                raise FrameTreeNameError(
                    outpt.field,
                    f"{outpt.field} is not in the output spec of '{self.name}' "
                    f"pipeline: " + "', '".join(field_names),
                )

    @property
    def input_varnames(self) -> ty.List[str]:
        return [
            i.name for i in self.inputs
        ]  # [path2varname(i.name) for i in self.inputs]

    @property
    def output_varnames(self) -> ty.List[str]:
        return [
            o.name for o in self.outputs
        ]  # [path2varname(o.name) for o in self.outputs]

    # parameterisation = self.get_parameterisation(kwargs)
    # self.wf.to_process.inputs.parameterisation = parameterisation
    # self.wf.per_node.source.inputs.parameterisation = parameterisation

    def __call__(self, ids: ty.List[str] = None) -> workflow.Task:
        """
        Create an "outer" workflow that interacts with the frameset to pull input
        data, process it and then push the derivatives back to the store.

        Parameters
        ----------
        **kwargs
            passed directly to the Pydra.Workflow init. The `ids` arg can be
            used to filter the data rows over which the pipeline is run.

        Returns
        -------
        pydra.compose.workflow.Task
            a Pydra workflow that iterates through the frameset, pulls data to the
            processing node, executes the analysis workflow on each data row,
            then uploads the outputs back to the data store

        Raises
        ------
        FrameTreeUsageError
            If the new pipeline will overwrite an existing pipeline connection
            with overwrite == False.
        """
        return PipelineWorkflow(
            task=self.task,
            ids=ids,
            frameset=self.frameset,
            row_frequency=self.row_frequency,
            inputs=self.inputs,
            outputs=self.outputs,
            converter_args=self.converter_args,
        )

    PROVENANCE_VERSION = "1.0"
    WORKFLOW_NAME = "processing"

    def asdict(
        self, required_modules: ty.Optional[ty.Set[str]] = None
    ) -> ty.Dict[str, ty.Any]:
        dct = asdict(self, omit=["task"], required_modules=required_modules)
        dct["task"] = pydra_asdict(self.task, required_modules=required_modules)
        return dct

    @classmethod
    def fromdict(cls, dct: ty.Dict[str, ty.Any], **kwargs: ty.Any) -> Self:
        return fromdict(dct, task=pydra_fromdict(dct["task"]), **kwargs)

    @classmethod
    def stack(
        cls, *sinks: ty.Union[SinkColumn, str]
    ) -> ty.List[ty.Tuple["Pipeline", ty.List[SinkColumn]]]:
        """Determines the pipelines stack, in order of execution,
        required to generate the specified sink columns.

        Parameters
        ----------
        sinks : Iterable[SinkColumn or str]
            the sink columns, or their names, that are to be generated

        Returns
        -------
        ty.List[tuple[Pipeline, ty.List[SinkColumn]]]
            stack of pipelines required to produce the specified data sinks,
            along with the sinks each stage needs to produce.

        Raises
        ------
        FrameTreeDesignError
            when there are circular references in the pipelines stack
        """

        # Stack of pipelines to process in reverse order of required execution
        stack = OrderedDict()

        def push_pipeline_on_stack(
            sink: SinkColumn, downstream: ty.Optional[ty.Tuple[Pipeline]] = None
        ) -> None:
            """
            Push a pipeline onto the stack of pipelines to be processed,
            detecting common upstream pipelines and resolving them to a single
            pipeline

            Parameters
            ----------
            sink: SinkColumn
                the sink to push its deriving pipeline for
            downstream : tuple[Pipeline]
                The pipelines directly downstream of the pipeline to be added.
                Used to detect circular dependencies
            """
            if downstream is None:
                downstream = []
            if sink.pipeline_name is None:
                raise FrameTreeDesignError(
                    f"{sink} hasn't been connected to a pipeline yet"
                )
            pipeline = sink.frameset.pipelines[sink.pipeline_name]
            if sink.name not in pipeline.output_varnames:
                raise FrameTreeOutputNotProducedException(
                    f"{pipeline.name} does not produce {sink.name}"
                )
            # Check downstream piplines for circular dependencies
            downstream_pipelines = [p for p, _ in downstream]
            if pipeline in downstream_pipelines:
                recur_index = downstream_pipelines.index(pipeline)
                raise FrameTreeDesignError(
                    f"{pipeline} cannot be a dependency of itself. Call-stack:\n"
                    + "\n".join(
                        "{} ({})".format(p, ", ".join(ro))
                        for p, ro in (
                            [[pipeline, sink.name]] + downstream[: (recur_index + 1)]
                        )
                    )
                )
            if pipeline.name in stack:
                # Pop pipeline from stack in order to add it to the end of the
                # stack and ensure it is run before all downstream pipelines
                prev_pipeline, to_produce = stack.pop(pipeline.name)
                assert pipeline is prev_pipeline
                # Combined required output to produce
                to_produce.append(sink)
            else:
                to_produce = []
            # Add the pipeline to the stack
            stack[pipeline.name] = pipeline, to_produce
            # Recursively add all the pipeline's prerequisite pipelines to the stack
            for inpt in pipeline.inputs:
                inpt_column = sink.frameset[inpt.name]
                if inpt_column.is_sink:
                    try:
                        push_pipeline_on_stack(
                            inpt_column,
                            downstream=[(pipeline, to_produce)] + downstream,
                        )
                    except FrameTreePipelinesStackError as e:
                        e.msg += (
                            "\nwhich are required as inputs to the '{}' "
                            "pipeline to produce '{}'".format(
                                pipeline.name, "', '".join(s.name for s in to_produce)
                            )
                        )
                        raise e

        # Add all pipelines
        for sink in sinks:
            push_pipeline_on_stack(sink)

        return reversed(stack.values())


def append_side_car_suffix(name: str, suffix: str) -> str:
    """Creates a new combined field name out of a basename and a side car"""
    return f"{name}__o__{suffix}"


def split_side_car_suffix(name: str) -> ty.List[str]:
    """Splits the basename from a side car sufix (as combined by `append_side_car_suffix`"""
    return name.split("__o__")


@workflow.define(outputs=["processed", "cant_process"])
def PipelineWorkflow(
    task: Task,
    frameset: FrameSet,
    row_frequency: Axes,
    inputs: ty.List[PipelineField],
    outputs: ty.List[PipelineField],
    converter_args: ty.Dict[str, dict],
    ids: ty.Optional[ty.List[str]] = None,
) -> ty.Tuple[ty.List[str], ty.List[str]]:
    """Create the outer workflow to link the analysis workflow with the
    data row iteration and store connection rows
    """

    # Generate list of rows to process checking existing outputs
    to_process = workflow.add(
        ToProcess(
            frameset=frameset,
            row_frequency=row_frequency,
            outputs=outputs,
            requested_ids=ids,
        )
    )

    per_row = workflow.add(
        PipelineRowWorkflow(
            task=task,
            frameset=frameset,
            row_frequency=row_frequency,
            inputs=inputs,
            outputs=outputs,
            converter_args=converter_args,
        ).split(row_id=to_process.row_ids)
    )

    return per_row.row_id, to_process.cant_process


@workflow.define(outputs=["row_id"])
def PipelineRowWorkflow(
    task: Task,
    frameset: frametree.core.frameset.base.FrameSet,
    row_frequency: Axes,
    row_id: str,
    inputs: ty.List[PipelineField],
    outputs: ty.List[PipelineField],
    converter_args: ty.Dict[str, dict],
):

    # Get the values from the frameset, caching remote files locally
    source = workflow.add(
        SourceItems(
            frameset=frameset,
            row_frequency=row_frequency,
            row_id=row_id,
            inputs=inputs,
        )
    )

    source_types = {}
    for inpt in inputs:
        if inpt.datatype is frametree.core.row.DataRow:
            # If the input datatype is a DataRow then the source is the whole
            # row
            source_types[inpt.name] = frametree.core.row.DataRow
            continue
        # If the row frequency of the column is not a parent of the pipeline
        # then the input will be a sequence of all the child rows
        dtype = frameset[inpt.name].datatype
        # If the row frequency of the source column is higher than the frequency
        # of the pipeline, then the related elements of the source column are
        # collected into a list and passed to the pipeline
        if inpt.datatype is not frametree.core.row.DataRow and not frameset[
            inpt.name
        ].row_frequency.is_parent(row_frequency, if_match=True):
            dtype = StateArray[dtype]
        source_types[inpt.name] = dtype

    column_names = list(source_types)

    # Dynamically access the inputs from the source dictionary
    @python.define(outputs=source_types)
    def SourceOutputs(
        sources: ty.Dict[
            str, ty.Union[frametree.core.row.DataRow, DataType, ty.List[DataType]]
        ],
        column_names: ty.List[str],
    ):
        return (
            sources[column_names[0]]
            if len(column_names) == 1
            else tuple(sources[c] for c in column_names)
        )

    source_outputs = workflow.add(
        SourceOutputs(sources=source.items, column_names=column_names)
    )

    # Set the inputs
    sourced = {i.name: getattr(source_outputs, i.name) for i in inputs}

    # Do input datatype conversions if required
    for inpt in inputs:
        if inpt.datatype == frametree.core.row.DataRow:
            continue
        stored_format = frameset[inpt.name].datatype
        if (
            inpt.datatype
            and not issubclass(inpt.datatype, stored_format)
            and not issubclass(stored_format, inpt.datatype)
        ):
            converter = inpt.datatype.get_converter(stored_format)
            logger.info(
                "Adding implicit conversion for input '%s' from %s to %s",
                inpt.name,
                stored_format.mime_like,
                inpt.datatype.mime_like,
            )
            in_file = sourced.pop(inpt.name)
            converter_task = copy(converter.task)
            for nm, val in converter_args.get(inpt.name, {}).items():
                setattr(converter_task, nm, val)
            if ty.get_origin(source_types[inpt.name]) is StateArray:
                # Iterate over all items in the sequence and convert them
                # separately
                converter_task.split(converter.in_file, **{converter.in_file: in_file})
            else:
                setattr(converter_task, converter.in_file, in_file)
            # Insert converter
            converter_outputs = workflow.add(
                converter_task, name=f"{inpt.name}_input_converter"
            )
            # Map converter output to input_interface
            sourced[inpt.name] = getattr(converter_outputs, converter.out_file)

    # Copy the task of the pipeline that actually performs the analysis/processing and
    # connections from the sourced/converted inputs
    task_copy = copy(task)

    for inpt in inputs:
        setattr(
            task_copy,
            inpt.field,
            sourced[inpt.name],
        )

    # Add the task to the workflow
    task_outputs = workflow.add(task_copy, name="task")

    # Set datatype converters where required
    to_sink = {o.name: getattr(task_outputs, o.field) for o in outputs}

    # Do output datatype conversions if required
    for outpt in outputs:
        stored_format = frameset[outpt.name].datatype
        sink_name = path2varname(outpt.name)
        if (
            outpt.datatype
            and not issubclass(outpt.datatype, stored_format)
            and not issubclass(stored_format, outpt.datatype)
        ):
            converter = stored_format.get_converter(outpt.datatype)
            logger.info(
                "Adding implicit conversion for output '%s' " "from %s to %s",
                outpt.name,
                outpt.datatype.mime_like,
                stored_format.mime_like,
            )
            # Insert converter
            task_kwargs = {converter.in_file: to_sink.pop(sink_name)}
            task_kwargs.update(converter_args.get(outpt.name, {}))
            converter_task = copy(converter.task)
            for nm, val in task_kwargs.items():
                setattr(converter_task, nm, val)
            converter_task = workflow.add(
                converter_task, name=f"{sink_name}_output_converter"
            )
            # Map converter output to workflow output
            to_sink[sink_name] = getattr(converter_task, converter.out_file)

    # Dynamically collect the outputs into a dictionary
    sink_types = {o.name: frameset[o.name].datatype for o in outputs}
    column_names = list(sink_types)

    @python.define(inputs=sink_types, outputs=["items"])
    def SinkInputs(
        column_names: ty.List[str],
        **sinks: DataType,
    ) -> ty.Dict[str, DataType]:
        return {c: sinks[c] for c in column_names}

    sink_inputs = workflow.add(SinkInputs(column_names=column_names, **to_sink))

    # Can't use a decorated function as we need to allow for dynamic
    # arguments

    sink = workflow.add(
        SinkItems(
            frameset=frameset,
            row_frequency=row_frequency,
            row_id=row_id,
            items=sink_inputs.items,
        )
    )
    # we just need to return something that can be connected to downstream nodes
    return sink.row_id


@python.define(outputs=["row_ids", "cant_process"])
def ToProcess(
    frameset: frametree.core.frameset.base.FrameSet,
    row_frequency: Axes,
    outputs: ty.List[PipelineField],
    requested_ids: ty.Union[ty.List[str], None],
) -> ty.Tuple[ty.List[str], ty.List[str]]:
    if requested_ids is None:
        requested_ids = frameset.row_ids(row_frequency)
    row_ids = []
    cant_process = []
    for row in frameset.rows(row_frequency, ids=requested_ids):
        # TODO: Should check provenance of existing rows to see if it matches
        empty = [row.cell(o.name).is_empty for o in outputs]
        if all(empty):
            row_ids.append(row.id)
        elif any(empty):
            cant_process.append(row.id)
    logger.debug(
        "Found %s ids to process, and can't process %s due to partially present outputs",
        row_ids,
        cant_process,
    )
    return row_ids, cant_process


@python.define(outputs=["items"])
def SourceItems(
    frameset: frametree.core.frameset.base.FrameSet,
    row_frequency: Axes,
    row_id: str,
    inputs: ty.List[PipelineField],
) -> ty.Dict[str, ty.Union["frametree.core.row.DataRow", DataType, ty.List[DataType]]]:
    """Selects the items from the frameset corresponding to the input
    sources and retrieves them from the store to a cache on
    the host

    Parameters
    ----------
    frameset : FrameSet
        the frameset to source the data from
    row_frequency : Axes
        the frequency of the row to source the data from
    row_id : str
        the ID of the row to source from

    Returns
    -------
    dict[str, DataType | DataRow]
        the sourced data items, or the whole row if the input
    """
    logger.debug("Sourcing %s", inputs)
    sourced: ty.Dict[str, ty.Union["frametree.core.row.DataRow", DataType]] = {}
    missing_inputs: ty.Dict[str, str] = {}
    row = frameset.row(row_frequency, row_id)
    with frameset.store.connection:
        for inpt in inputs:
            # If the required datatype is of type DataRow then provide the whole
            # row to the pipeline input
            if inpt.datatype == frametree.core.row.DataRow:
                sourced[inpt.name] = row
                continue
            try:
                sourced[inpt.name] = row[inpt.name]
            except FrameTreeDataMatchError as e:
                missing_inputs[inpt.name] = str(e)
    if missing_inputs:
        raise FrameTreeDataMatchError("\n\n" + "\n\n".join(missing_inputs.values()))
    return sourced


@python.define(outputs=["row_id"])
def SinkItems(
    frameset: FrameSet,
    row_frequency: Axes,
    row_id: str,
    items: ty.Dict[str, ty.Any],
    provenance: ty.Optional[ty.Dict[str, ty.Any]] = None,
) -> str:
    """Stores items generated by the pipeline back into the store

    Parameters
    ----------
    frameset : FrameSet
        the frameset to source the data from
    row_frequency : Axes
        the frequency of the row to source the data from
    row_id : str
        the ID of the row to source from
    provenance : dict
        provenance information to be stored alongside the generated data
    **to_sink : dict[str, DataType]
        data items to be stored in the data store

    Returns
    -------
    row_id: str
        the ID of the row that was processed
    """
    if provenance is not None:
        raise NotImplementedError("Provenance storage not implemented yet")
    logger.debug("Sinking %s", items)
    row = frameset.row(row_frequency, row_id)
    with frameset.store.connection:
        for outpt_name, output in items.items():
            row.cell(outpt_name).item = output
    return row_id


# Provenance mismatch detection methods salvaged from data.provenance

# def mismatches(self, other, include=None, exclude=None):
#     """
#     Compares information stored within provenance objects with the
#     exception of version information to see if they match. Matches are
#     constrained to the name_paths passed to the 'include' kwarg, with the
#     exception of sub-name_paths passed to the 'exclude' kwarg

#     Parameters
#     ----------
#     other : Provenance
#         The provenance object to compare against
#     include : list[ty.List[str]] | None
#         Paths in the provenance to include in the match. If None all are
#         incluced
#     exclude : list[ty.List[str]] | None
#         Paths in the provenance to exclude from the match. In None all are
#         excluded
#     """
#     if include is not None:
#         include_res = [self._gen_prov_path_regex(p) for p in include]
#     if exclude is not None:
#         exclude_res = [self._gen_prov_path_regex(p) for p in exclude]
#     diff = DeepDiff(self._prov, other._prov, ignore_order=True)
#     # Create regular expressions for the include and exclude name_paths in
#     # the datatype that deepdiff uses for nested dictionary/lists

#     def include_change(change):
#         if include is None:
#             included = True
#         else:
#             included = any(rx.match(change) for rx in include_res)
#         if included and exclude is not None:
#             included = not any(rx.match(change) for rx in exclude_res)
#         return included

#     filtered_diff = {}
#     for change_type, changes in diff.items():
#         if isinstance(changes, dict):
#             filtered = dict((k, v) for k, v in changes.items()
#                             if include_change(k))
#         else:
#             filtered = [c for c in changes if include_change(c)]
#         if filtered:
#             filtered_diff[change_type] = filtered
#     return filtered_diff

# @classmethod
# def _gen_prov_path_regex(self, file_path):
#     if isinstance(file_path, str):
#         if file_path.startswith('/'):
#             file_path = file_path[1:]
#         regex = re.compile(r"root\['{}'\].*"
#                             .format(r"'\]\['".join(file_path.split('/'))))
#     elif not isinstance(file_path, re.Pattern):
#         raise FrameTreeUsageError(
#             "Provenance in/exclude name_paths can either be name_path "
#             "strings or regexes, not '{}'".format(file_path))
#     return regex



class FrameTreeException(Exception):
    @property
    def msg(self):
        return self.args[0]

    @msg.setter
    def msg(self, msg):
        self.args = (msg,) + self.args[1:]


class FrametreeCannotSerializeDynamicDefinitionError(FrameTreeException):
    """
    Raised when a FrameTree object cannot be serialized
    """

    def __init__(self, msg: str):
        super().__init__(msg)
        self.msg = msg


class FrameTreeError(FrameTreeException):
    pass


class FrameTreeRuntimeError(FrameTreeError):
    pass


class FrameTreeNotBoundToAnalysisError(FrameTreeError):
    pass


class FrameTreeVersionError(FrameTreeError):
    pass


class FrameTreeRequirementNotFoundError(FrameTreeVersionError):
    pass


class FrameTreeVersionNotDetectableError(FrameTreeVersionError):
    pass


class FrameTreeEnvModuleNotLoadedError(FrameTreeError):
    pass


class FrameTreeMissingInputError(FrameTreeException):
    pass


class FrameTreeProtectedOutputConflictError(FrameTreeError):
    pass


class FrameTreeCantPickleAnalysisError(FrameTreeError):
    pass


class FrameTreeRepositoryError(FrameTreeError):
    pass


class FrameTreeUsageError(FrameTreeError):
    pass


class FrameTreeCacheError(FrameTreeError):
    pass


class FrameTreeDesignError(FrameTreeError):
    pass


class NamedFrameTreeError(FrameTreeError):
    def __init__(self, name: str, msg: str) -> None:
        super(NamedFrameTreeError, self).__init__(msg)
        self.name = name


class FrameTreeNameError(NamedFrameTreeError):
    pass


class FrameTreeWrongFrequencyError(NamedFrameTreeError):
    pass


class FrameTreeIndexError(FrameTreeError):
    def __init__(self, index: int, msg: str) -> None:
        super(FrameTreeIndexError, self).__init__(msg)
        self.index = index


class FrameTreeDataMatchError(FrameTreeUsageError):
    pass


class FrameTreePipelinesStackError(FrameTreeError):
    pass


class FrameTreeMissingDataException(FrameTreePipelinesStackError):
    pass


class FrameTreeOutputNotProducedException(FrameTreePipelinesStackError):
    """
    Raised when a given spec is not produced due to switches and inputs
    provided to the analysis
    """


class FrameTreeInsufficientRepoDepthError(FrameTreeError):
    pass


class FrameTreeLicenseNotFoundError(FrameTreeNameError):
    pass


class FrameTreeUnresolvableFormatException(FrameTreeException):
    pass


class FrameTreeFileSetNotCachedException(FrameTreeException):
    pass


class NoMatchingPipelineException(FrameTreeException):
    pass


class FrameTreeModulesError(FrameTreeError):
    pass


class FrameTreeModulesNotInstalledException(FrameTreeException):
    pass


class FrameTreeJobSubmittedException(FrameTreeException):
    """
    Signifies that a pipeline has been submitted to a scheduler and
    a return value won't be returned.
    """


class FrameTreeNoRunRequiredException(FrameTreeException):
    """
    Used to signify when a pipeline doesn't need to be run as all
    required outputs are already present in the store
    """


class FrameTreeFileFormatClashError(FrameTreeError):
    """
    Used when two mismatching data formats are registered with the same
    name or extension
    """


class FrameTreeConverterNotAvailableError(FrameTreeError):
    "The converter required to convert between formats is not"

    "available"


class FrameTreeReprocessException(FrameTreeException):
    pass


class FrameTreeWrongRepositoryError(FrameTreeError):
    pass


class FrameTreeIvalidParameterError(FrameTreeError):
    pass


class FrameTreeRequirementVersionsError(FrameTreeError):
    pass


class FrameTreeXnatCommandError(FrameTreeRepositoryError):
    """
    Error in the command file used to access an XNAT repository via the XNAT
    container service.
    """


class FrameTreeUriAlreadySetException(FrameTreeException):
    """Raised when attempting to set the URI of an item is already set"""


class FrameTreeConstructionError(FrameTreeError):
    "Error in constructing data tree by store find_rows method"


class FrameTreeBadlyFormattedIDError(FrameTreeConstructionError):
    "Error attempting to extract an ID from a tree path using a user provided regex"


class FrameTreeWrongAxesError(FrameTreeError):
    "Provided row_frequency is not a valid member of the dataset's dimensions"


class FrameTreeNoDirectXnatMountException(FrameTreeException):
    "Raised when attemptint to access a file-system mount for a row that hasn't been mounted directly"

    pass


class FrameTreeEmptyDatasetError(FrameTreeException):
    pass


class FrameTreeBuildError(FrameTreeError):
    pass


class NamedError(Exception):
    def __init__(self, name, msg):
        super().__init__(msg)
        self.name = name


class NameError(NamedError):
    pass


class DataNotDerivedYetError(NamedError):
    pass


class DatatypeUnsupportedByStoreError(FrameTreeError):
    """Raised when a data store doesn't support a given datatype"""

    def __init__(self, datatype, store):
        super().__init__(
            f"'{datatype.mime_like}' data types aren't supported by {type(store)} stores"
        )



from operator import mul
from pathlib import Path
from functools import reduce
from fileformats.core import FileSet
from fileformats.text import TextFile
from fileformats.field import Text, Boolean
from frametree.common import FileSystem
from frametree.testing.blueprint import (
    TestDatasetBlueprint,
    TestAxes,
    FileSetEntryBlueprint,
    FieldEntryBlueprint,
)
from frametree.core.frameset.base import FrameSet


def test_column_api_access(dataset: FrameSet):

    bp = dataset.__annotations__["blueprint"]

    for fileset_bp in bp.entries:

        dataset.add_source(fileset_bp.path, fileset_bp.datatype)

        col = dataset[fileset_bp.path]

        # Check length of column
        assert len(col) == reduce(mul, bp.dim_lengths)

        # Access file-set via leaf IDs
        with dataset.tree:
            for id_ in col.ids:
                item = col[id_]
                assert isinstance(item, fileset_bp.datatype)
                if issubclass(fileset_bp.datatype, FileSet):
                    assert sorted(p.name for p in item.fspaths) == sorted(
                        fileset_bp.filenames
                    )


def test_column_datatype_conversion(tmp_path: Path):
    # Check that the datatype conversion works
    frameset = TestDatasetBlueprint(  # dataset name
        axes=TestAxes,
        hierarchy=["a", "b", "c", "d"],
        dim_lengths=[1, 1, 1, 1],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=TextFile, filenames=["file.txt"]
            ),
            FieldEntryBlueprint(
                path="textfield",
                row_frequency="abcd",
                datatype=Text,
                value="sample-text",
            ),  # Derivatives to insert
            FieldEntryBlueprint(
                path="booleanfield",
                row_frequency="c",
                datatype=Boolean,
                value="no",
                expected_value=False,
            ),  # Derivatives to insert
        ],
    ).make_dataset(
        store=FileSystem(),
        dataset_id=tmp_path / "field-datatype-conversion",
        name="",
    )  # type: ignore[no-any-return]

    ID_KEY = ("a0", "b0", "c0", "d0")

    frameset.add_source("text_file", "text/text-file", path="file1")
    frameset.add_source("text_field", str, path="textfield")
    frameset.add_source("boolean_field", bool, path="booleanfield")
    assert [p.name for p in frameset["text_file"][ID_KEY].fspaths] == ["file1.txt"]
    assert frameset["text_field"][ID_KEY].value == "sample-text"
    assert frameset["boolean_field"][ID_KEY].value is False



import typing as ty
import json
from pydra.compose import python
from fileformats.generic import FileSet, File


def identity(**fields: ty.Any) -> ty.Tuple[ty.Any, ...]:
    return tuple(fields.values())


def define_identity(fields: ty.List[str]) -> ty.Callable:
    return python.define(identity, inputs=fields, outputs=fields)


@python.define
def IdentityConverter(in_file: FileSet) -> FileSet:
    """Returns the input file set

    Parameters
    ----------
    in_file : FileSet
        The input file set

    Returns
    -------
    out_file: FileSet
        The input file set
    """
    return in_file


@python.define
def ExtractFromJson(in_file: File, field_name: str) -> ty.Any:
    with open(in_file) as f:
        dct = json.load(f)
    return dct[field_name]  # FIXME: Should use JSONpath syntax



from pathlib import Path
import cloudpickle as cp
from pydra.compose import python, workflow
from pydra.engine.job import Job
from pydra.engine.submitter import Submitter
from pydra.utils.hash import hash_object
from frametree.core.frameset.base import FrameSet
from frametree.core.serialize import asdict, fromdict


def test_dataset_asdict_roundtrip(dataset):

    dct = asdict(dataset, omit=["store", "id"])
    undct = fromdict(dct, store=dataset.store, id=dataset.id)
    assert isinstance(dct, dict)
    assert "store" not in dct
    del dataset.__annotations__["blueprint"]
    assert dataset == undct


def test_dataset_pickle(dataset: FrameSet, tmp_dir: Path):
    fpath = tmp_dir / "dataset.pkl"
    with fpath.open("wb") as fp:
        cp.dump(dataset, fp)
    with fpath.open("rb") as fp:
        reloaded = cp.load(fp)
    assert dataset == reloaded


def test_dataset_in_workflow_pickle(dataset: FrameSet, tmp_dir: Path):

    # Create the outer workflow to link the analysis workflow with the
    # data row iteration and store connection rows
    @workflow.define(outputs={"c": int})
    def Workflow(a):

        test_func = workflow.add(func(a=wf.lzin.a, b=2, dataset=dataset))

        return test_func.c

    job = Job(
        task=Workflow(a=1),
        submitter=Submitter(),
        name="job",
    )

    job.pickle_task()


@python.define(outputs=["c"])
def func(a: int, b: int, dataset: FrameSet) -> int:
    return a + b


def test_dataset_bytes_hash(dataset):

    hsh = hash_object(dataset)
    # Check hashing is stable
    assert hash_object(dataset) == hsh



from urllib.parse import urlparse
from pathlib import Path, PurePath
import attrs


@attrs.define
class License:
    """Specification of a software license that needs to be present in the container
    when the command is run.

    Parameters
    ----------
    name : str
        a name to refer to the license with. Must be unique among the licenses used
        pipelines applied to a dataset, ideally for a site. Typically named closely
        after the package it is used for along with a version number if the license,
        needs to be updated with new versions e.g. freesurfer, fsl, matlab_v2022a etc...
    destination : PurePath
        destination within the container to install the license
    description : str
        a short description of the license and what it is used for
    info_url : str
        link to website with information about license, particularly how to download it
    source : Path, optional
        path to the location of a valid license file
    store_in_image : bool
        whether the license can be stored in the image or not according to the license
        conditions
    """

    name: str = attrs.field()
    destination: PurePath = attrs.field(converter=PurePath)
    description: str = attrs.field()
    info_url: str = attrs.field()
    source: Path = attrs.field(
        default=None, converter=lambda x: Path(x) if x is not None else None
    )
    store_in_image: bool = False

    @info_url.validator
    def info_url_validator(self, _, info_url):
        parsed = urlparse(info_url)
        if not parsed.scheme or not parsed.netloc:
            raise ValueError(
                f"Could not parse info url '{info_url}', please include URL scheme"
            )

    # FIXME: this doesn't work inside images
    # @source.validator
    # def source_validator(self, _, source):
    #     if source is not None and not source.exists():
    #         raise ValueError(
    #             f"Source file for {self.name} license, '{str(source)}', does not exist"
    #         )

    @classmethod
    def column_path(self, name):
        """The column name (and resource name) for the license if it is to be downloaded
        from the source dataset"""
        return name + self.COLUMN_SUFFIX + "@"

    COLUMN_SUFFIX = "_LICENSE"



from pathlib import Path
import logging
import typing as ty
import cloudpickle as cp
import click
import pydra.compose.base
from fileformats.core import from_mime
from frametree.core.frameset.base import FrameSet
from frametree.core.store import Store
from frametree.core.serialize import ClassResolver, parse_value
from frametree.core.utils import set_loggers
from .base import cli


logger = logging.getLogger("frametree")


@cli.command(
    name="derive",
    help="""Derive data for a data sink column and all prerequisite columns.

ADDRESS string containing the nickname of the data store, the ID of the dataset
(e.g. XNAT project ID or file-system directory) and the dataset's name in the
format <store-nickname>//<dataset-id>[@<dataset-name>]

COLUMNS are the names of the sink columns to derive""",
)
@click.argument("address")
@click.argument("columns", nargs=-1)
@click.option(
    "--work",
    "-w",
    default=None,
    help=(
        "The location of the directory where the working files "
        "created during the pipeline execution will be stored"
    ),
)
@click.option(
    "--plugin",
    default="cf",
    help=("The Pydra plugin with which to process the workflow"),
)
@click.option(
    "--loglevel",
    type=str,
    default="info",
    help=("The level of detail logging information is presented"),
)
def derive(address, columns, work, plugin, loglevel):

    logging.basicConfig(level=getattr(logging, loglevel.upper()))

    if work is not None:
        work_dir = Path(work)
        store_cache = work_dir / "store-cache"
        pipeline_cache = work_dir / "pipeline-cache"
    else:
        store_cache = None
        pipeline_cache = None

    dataset = FrameSet.load(address, cache_dir=store_cache)

    set_loggers(loglevel)

    dataset.derive(*columns, cache_dir=pipeline_cache, worker=plugin)

    columns_str = "', '".join(columns)
    logger.info(f"Derived data for '{columns_str}' column(s) successfully")


@cli.command(help="""Display the potential derivatives that can be derived""")
def menu():
    raise NotImplementedError


@cli.command(
    name="show-errors",
    help="""Show a Pydra crash report

NODE_WORK_DIR is the directory containing the error pickle file""",
)
@click.argument("node_work_dir")
def show_errors(node_work_dir):
    node_work_dir = Path(node_work_dir)
    files = ["_task.pklz", "_result.pklz", "_error.pklz"]  #
    for fname in files:
        fpath = node_work_dir / fname
        if fpath.exists():
            with open(fpath, "rb") as f:
                contents = cp.load(f)
            click.echo(f"{fname}:")
            if isinstance(contents, dict):
                for k, v in contents.items():
                    if k == "error message":
                        click.echo(f"{k}:\n" + "".join(v))
                    else:
                        click.echo(f"{k}: {v}")
            else:
                click.echo(contents)


@cli.command(
    name="ignore-diff",
    help="""Ignore difference between provenance of previously generated derivative
and new parameterisation""",
)
def ignore_diff():
    raise NotImplementedError


if __name__ == "__main__":
    from click.testing import CliRunner

    runner = CliRunner()
    runner.invoke(
        show_errors,
        [
            (
                "/Users/tclose/Downloads/892d1907-fe2b-40ac-9b77-a3f2ee21ca76/pipeline-cache/"
                "Workflow_f3a2bb7474848840aec86fd87e88f5938217fae5976fc699bfc993d95d48a3b8"
            )
        ],
        catch_exceptions=False,
    )


@cli.command(
    name="apply",
    help="""Apply a Pydra workflow to a dataset as a pipeline between
two columns

ADDRESS string containing the nickname of the data store, the ID of the dataset
(e.g. XNAT project ID or file-system directory) and the dataset's name in the
format <store-nickname>//<dataset-id>[@<dataset-name>]

PIPELINE_NAME is the name of the pipeline

WORKFLOW_LOCATION is the location to a Pydra workflow on the Python system path,
<MODULE>:<WORKFLOW>""",
)
@click.argument("address")
@click.argument("pipeline_name")
@click.argument("workflow_location")
@click.option(
    "--input",
    "-i",
    nargs=3,
    default=(),
    metavar="<col-name> <pydra-field> <required-datatype>",
    multiple=True,
    type=str,
    help=(
        "the link between a column and an input of the workflow. "
        "The required format is the location (<module-path>:<class>) of the format "
        "expected by the workflow"
    ),
)
@click.option(
    "--output",
    "-o",
    nargs=3,
    default=(),
    metavar="<col-name> <pydra-field> <produced-datatype>",
    multiple=True,
    type=str,
    help=(
        "the link between an output of the workflow and a sink column. "
        "The produced datatype is the location (<module-path>:<class>) of the datatype "
        "produced by the workflow"
    ),
)
@click.option(
    "--parameter",
    "-p",
    nargs=2,
    default=(),
    metavar="<name> <value>",
    multiple=True,
    type=str,
    help=("a fixed parameter of the workflow to set when applying it"),
)
@click.option(
    "--source",
    "-s",
    nargs=3,
    default=(),
    metavar="<col-name> <pydra-field> <required-datatype>",
    multiple=True,
    type=str,
    help=(
        "add a source to the dataset and link it to an input of the workflow "
        "in a single step. The source column must be able to be specified by its "
        "path alone and be already in the datatype required by the workflow"
    ),
)
@click.option(
    "--sink",
    "-k",
    nargs=3,
    default=(),
    metavar="<col-name> <pydra-field> <produced-datatype>",
    multiple=True,
    type=str,
    help=(
        "add a sink to the dataset and link it to an output of the workflow "
        "in a single step. The sink column be in the same datatype as produced "
        "by the workflow"
    ),
)
@click.option(
    "--row-frequency",
    "-f",
    default=None,
    type=str,
    help=(
        "the row-frequency of the rows the pipeline will be executed over, i.e. "
        "will it be run once per-session, per-subject or per whole dataset, "
        "by default the highest row-frequency rows (e.g. per-session)"
    ),
)
@click.option(
    "--overwrite/--no-overwrite",
    default=False,
    help=("whether to overwrite previous pipelines"),
)
def apply(
    address,
    pipeline_name,
    workflow_location,
    input,
    output,
    parameter,
    source,
    sink,
    row_frequency,
    overwrite,
):

    frameset = FrameSet.load(address)
    workflow = ClassResolver(pydra.compose.base.Task, alternative_types=[ty.Callable])(
        workflow_location
    )(**{n: parse_value(v) for n, v in parameter})

    inputs = parse_col_option(input)
    outputs = parse_col_option(output)
    sources = parse_col_option(source)
    sinks = parse_col_option(sink)

    for col_name, field, datatype in sources:
        frameset.add_source(col_name, datatype)
        inputs.append((col_name, field, datatype))

    for col_name, field, datatype in sinks:
        frameset.add_sink(col_name, datatype)
        outputs.append((col_name, field, datatype))

    frameset.apply(
        pipeline_name,
        workflow,
        inputs,
        outputs,
        row_frequency=row_frequency,
        overwrite=overwrite,
    )

    frameset.save()


def parse_col_option(option):
    return [(c, p, from_mime(f)) for c, p, f in option]


@cli.command(
    name="install-license",
    help="""Installs a license within a store (i.e. site-wide) or dataset (project-specific)
for use in a deployment pipeline

LICENSE_NAME the name of the license to upload. Must match the name of the license specified
in the deployment specification

SOURCE_FILE path to the license file to upload

INSTALL_LOCATIONS a list of installation locations, which are either the "nickname" of a
store (as saved by `frametree store add`) or the ID of a dataset in form
<store-nickname>//<dataset-id>[@<dataset-name>], where the dataset ID
is either the location of the root directory (for file-system based stores) or the
project ID for managed data repositories.
""",
)
@click.argument("license_name")
@click.argument("source_file", type=click.Path(exists=True, path_type=Path))
@click.argument("install_locations", nargs=-1)
@click.option(
    "--logfile",
    default=None,
    type=click.Path(path_type=Path),
    help="Log output to file instead of stdout",
)
@click.option("--loglevel", default="info", help="The level to display logs at")
def install_license(install_locations, license_name, source_file, logfile, loglevel):
    logging.basicConfig(filename=logfile, level=getattr(logging, loglevel.upper()))

    if isinstance(source_file, bytes):  # FIXME: This shouldn't be necessary
        source_file = Path(source_file.decode("utf-8"))

    if not install_locations:
        install_locations = ["file_system"]

    for install_loc in install_locations:
        if "//" in install_loc:
            dataset = FrameSet.load(install_loc)
            store_name, _, _ = FrameSet.parse_id_str(install_loc)
            msg = f"for '{dataset.name}' dataset on {store_name} store"
        else:
            store = Store.load(install_loc)
            dataset = store.site_licenses_dataset()
            if dataset is None:
                raise ValueError(
                    f"{install_loc} store doesn't support the installation of site-wide "
                    "licenses, please specify a dataset to install it for"
                )
            msg = f"site-wide on {install_loc} store"

        dataset.install_license(license_name, source_file)
        logger.info("Successfully installed '%s' license %s", license_name, msg)



from __future__ import annotations
import os
from pathlib import Path
import re
import typing as ty
import logging
import json
import attrs
from fileformats.core import FileSet, Field
from frametree.core.exceptions import FrameTreeUsageError
from frametree.core.frameset.base import DataTree
from frametree.core.axes import Axes
from frametree.core.row import DataRow
from frametree.core.entry import DataEntry
from frametree.core.store import LocalStore
from frametree.core.utils import full_path


logger = logging.getLogger("frametree")


# Matches directory names used for summary rows with dunder beginning and
# end (e.g. '__visit_01__') and hidden directories (i.e. starting with '.' or
# '~')
special_dir_re = re.compile(r"(__.*__$|\..*|~.*)")


@attrs.define
class FileSystem(LocalStore):
    """
    A Repository class for data stored hierarchically within sub-directories
    of a file-system directory. The depth and which layer in the data tree
    the sub-directories correspond to is defined by the `hierarchy` argument.
    """

    PROV_SUFFIX = ".provenance"
    FIELDS_FNAME = "__fields__.json"
    FIELDS_PROV_FNAME = "__fields_provenance__.json"

    # Note this name will be constant, as there is only ever one store,
    # which covers whole FS
    name: str = "file_system"

    #################################
    # Abstract-method implementations
    #################################

    def populate_tree(self, tree: DataTree) -> None:
        """
        Scans the data present in the dataset and populates the data tree with nodes

        Parameters
        ----------
        dataset : FrameSet
            The dataset to construct the tree dimensions for
        """
        if not os.path.exists(tree.dataset_id):
            raise FrameTreeUsageError(
                f"Could not find a directory at '{tree.dataset_id}' to be the "
                "root row of the dataset"
            )
        dpaths = sorted(d for d, _, _ in os.walk(tree.dataset_id))
        for dpath in dpaths:
            tree_path = tuple(Path(dpath).relative_to(tree.dataset_id).parts)
            if len(tree_path) != len(tree.hierarchy):
                continue
            if self.FRAMETREE_DIR in tree_path:
                continue
            tree.add_leaf(tree_path)

    def populate_row(self, row: DataRow) -> None:
        """Scans the node in the data tree corresponding to the data row and populates
        the row with data entries found in the tree node

        Parameters
        ----------
        row : DataRow
            the data row to populate
        """

        def filter_entry_dir(entry_dir: Path) -> ty.Iterator[Path]:
            for subpath in entry_dir.iterdir():
                entry_name = subpath.name
                if (
                    not entry_name.startswith(".")
                    and entry_name != self.FRAMETREE_DIR
                    and entry_name
                    not in (
                        self.FIELDS_FNAME,
                        self.FIELDS_PROV_FNAME,
                        self.FIELDS_FNAME + self.LOCK_SUFFIX,
                    )
                    and not entry_name.endswith(self.PROV_SUFFIX)
                ):
                    yield subpath

        root_dir = full_path(row.frameset.id)

        # Iterate through all directories saved for the source and dataset derivatives
        for dataset_name in self._row_dataset_names(row):
            row_dir = root_dir / self._row_relpath(row, dataset_name=dataset_name)
            if row_dir.exists():
                # Filter contents of directory to omit fields JSON and provenance and
                # add file-set entries
                for entry_fspath in filter_entry_dir(row_dir):
                    path = str(entry_fspath.relative_to(row_dir))
                    if dataset_name is not None:
                        path += "@" + dataset_name
                    row.add_entry(
                        path=path,
                        datatype=FileSet,
                        uri=str(entry_fspath.relative_to(root_dir)),
                    )
                # Add field entries
                fields_json = row_dir / self.FIELDS_FNAME
                try:
                    with open(fields_json) as f:
                        fields_dict = json.load(f)
                except FileNotFoundError:
                    pass
                else:
                    for name in fields_dict:
                        path = (
                            f"{name}@{dataset_name}"
                            if dataset_name is not None
                            else name
                        )
                        row.add_entry(
                            path=path,
                            datatype=Field,
                            uri=str(fields_json.relative_to(root_dir)) + "::" + name,
                        )

    def get_field(
        self, entry: DataEntry, datatype: ty.Type[Field[ty.Any, ty.Any]]
    ) -> Field[ty.Any, ty.Any]:
        """Retrieve the field associated with the given entry and return it cast
        to the specified datatype

        Parameters
        ----------
        entry : DataEntry
            the entry to retrieve the field for
        datatype : type (subclass DataType)
            the datatype to return the field as

        Returns
        -------
        Field
            the retrieved field
        """
        fspath, key = self._fields_fspath_and_key(entry)
        return datatype(self.read_from_json(fspath, key))

    def get_fileset(self, entry: DataEntry, datatype: ty.Type[FileSet]) -> FileSet:
        """Retrieve the file-set associated with the given entry and return it cast
        to the specified datatype

        Parameters
        ----------
        entry : DataEntry
            the entry to retrieve the file-set for
        datatype : type (subclass DataType)
            the datatype to return the file-set as

        Returns
        -------
        FileSet
            the retrieved file-set
        """
        return datatype(self._fileset_fspath(entry))

    def put_fileset(self, fileset: FileSet, entry: DataEntry) -> FileSet:
        """Put a file-set into the specified data entry

        Parameters
        ----------
        fileset : FileSet
            the file-set to store
        entry : DataEntry
            the entry to store the file-set in

        Returns
        -------
        FileSet
            the copy of the file-set that has been stored within the data entry
        """
        fspath = self._fileset_fspath(entry)
        _, new_stem, new_ext = FileSet.decompose_fspath(entry.path.split("@")[0])
        if new_ext:
            if len(fileset.fspaths) > 1:
                raise FrameTreeUsageError(
                    "Cannot store file-set with multiple files in file_system store "
                    "when extension is specified"
                )
            if new_ext != FileSet.decompose_fspath(fileset.fspath)[2]:
                raise FrameTreeUsageError(
                    "Cannot change extension of file-set when copying to file_system store"
                )
        # Create target directory if it doesn't exist already
        copied_fileset = fileset.copy(
            dest_dir=fspath.parent,
            collation=fileset.CopyCollation.adjacent,
            new_stem=new_stem,
            make_dirs=True,
            overwrite=True,
        )
        return copied_fileset

    def put_field(self, field: Field, entry: DataEntry) -> None:
        """Put a field into the specified data entry

        Parameters
        ----------
        field : Field
            the field to store
        entry : DataEntry
            the entry to store the field in
        """
        fspath, key = self._fields_fspath_and_key(entry)
        self.update_json(fspath, key, field.primitive(field))

    def get_fileset_provenance(
        self, entry: DataEntry
    ) -> ty.Union[ty.Dict[str, ty.Any], None]:
        """Retrieves provenance associated with a file-set data entry

        Parameters
        ----------
        entry : DataEntry
            the entry of the file-set to retrieve the provenance for

        Returns
        -------
        ty.Dict[str, ty.Any] or None
            the retrieved provenance or None if it doesn't exist
        """
        with open(self._fileset_prov_fspath(entry)) as f:
            provenance = json.load(f)
        return provenance

    def put_fileset_provenance(
        self, provenance: ty.Dict[str, ty.Any], entry: DataEntry
    ) -> None:
        """Puts provenance associated with a file-set data entry into the store

        Parameters
        ----------
        provenance : dict[str, ty.Any]
            the provenance to store
        entry : DataEntry
            the entry to associate the provenance with
        """
        with open(self._fileset_prov_fspath(entry), "w") as f:
            json.dump(provenance, f)

    def get_field_provenance(
        self, entry: DataEntry
    ) -> ty.Union[ty.Dict[str, ty.Any], None]:
        """Retrieves provenance associated with a field data entry

        Parameters
        ----------
        entry : DataEntry
            the entry of the field to retrieve the provenance for

        Returns
        -------
        ty.Dict[str, ty.Any] or None
            the retrieved provenance or None if it doesn't exist
        """
        fspath, key = self._fields_prov_fspath_and_key(entry)
        with open(fspath) as f:
            fields_provenance = json.load(f)
        return fields_provenance[key]

    def put_field_provenance(
        self, provenance: ty.Dict[str, ty.Any], entry: DataEntry
    ) -> None:
        """Puts provenance associated with a field data entry into the store

        Parameters
        ----------
        provenance : dict[str, ty.Any]
            the provenance to store
        entry : DataEntry
            the entry to associate the provenance with
        """
        fspath, key = self._fields_prov_fspath_and_key(entry)
        self.update_json(fspath, key, provenance)

    def fileset_uri(self, path: str, datatype: type, row: DataRow) -> str:
        """Returns the "uri" (e.g. file-system path relative to root dir) of a file-set
        entry at the given path relative to the given row

        Parameters
        ----------
        path : str
            path to the entry relative to the row
        datatype : type
            the datatype of the entry
        row : DataRow
            the row of the entry

        Returns
        -------
        uri : str
            the "uri" to the file-set entry relative to the data store
        """
        path, dataset_name = DataEntry.split_dataset_name_from_path(path)
        row_dir = self._row_relpath(row, dataset_name=dataset_name)
        return str(row_dir.joinpath(*path.split("/"))) + datatype.strext

    def field_uri(self, path: str, datatype: type, row: DataRow) -> str:
        """Returns the "uri" (e.g. file-system path relative to root dir) of a field
        entry at the given path relative to the given row

        Parameters
        ----------
        path : str
            path to the entry relative to the row
        datatype : type
            the datatype of the entry
        row : DataRow
            the row of the entry

        Returns
        -------
        uri : str
            the "uri" to the field entry relative to the data store
        """
        path, dataset_name = DataEntry.split_dataset_name_from_path(path)
        row_dir = self._row_relpath(row, dataset_name=dataset_name)
        return str(row_dir / self.FIELDS_FNAME) + "::" + path

    def create_data_tree(
        self,
        id: str,
        leaves: ty.List[ty.Tuple[str, ...]],
        hierarchy: ty.List[str],
        axes: ty.Type[Axes],
        **kwargs: ty.Any,
    ) -> None:
        """creates a new empty dataset within in the store. Used in test routines and
        importing/exporting datasets between stores

        Parameters
        ----------
        id : str
            ID for the newly created dataset
        leaves : list[tuple[str, ...]]
                        list of IDs for each leaf node to be added to the dataset. The IDs for each
            leaf should be a tuple with an ID for each level in the tree's hierarchy, e.g.
            for a hierarchy of [subject, visit] ->
            [("SUBJ01", "TIMEPOINT01"), ("SUBJ01", "TIMEPOINT02"), ....]
        hierarchy: ty.List[str]
            the hierarchy of the dataset to be created
        space : type(Axes)
            the data space of the dataset
        """
        root_dir = Path(id)
        root_dir.mkdir(parents=True)
        # Create sub-directories corresponding to rows of the dataset
        for ids_tuple in leaves:
            root_dir.joinpath(*ids_tuple).mkdir(parents=True)

    ##################
    # Helper functions
    ##################

    def _row_relpath(self, row: DataRow, dataset_name: ty.Optional[str] = None) -> Path:
        """Get the file-system path to the dataset root for the given row, taking into
        account non-leaf rows

        Parameters
        ----------
        row : DataRow
            the row to get the relative path for
        derivatives : bool
            whether to return the directory containing derivatives or originals

        Returns
        -------
        relpath : Path
            the relative path to the row directory
        """
        relpath = Path()
        if row.frequency is max(row.frameset.axes):  # leaf node
            for freq in row.frameset.hierarchy:
                relpath /= row.frequency_id(freq)
            if dataset_name is not None:
                relpath /= self.FRAMETREE_DIR
                if dataset_name:
                    relpath /= dataset_name
                else:
                    relpath /= self.EMPTY_DATASET_NAME
        else:
            relpath = relpath.joinpath(
                self.FRAMETREE_DIR,
                str(row.frequency),
            )
            if isinstance(row.id, tuple):
                relpath /= ".".join(row.id)
            elif row.id:
                relpath /= row.id
            if dataset_name is None:
                relpath /= self.FRAMETREE_DIR
            elif not dataset_name:
                relpath /= self.EMPTY_DATASET_NAME
            else:
                relpath /= dataset_name
        return relpath

    def _row_dataset_names(self, row: DataRow) -> ty.List[str]:
        """list all dataset names stored in the given row

        Parameters
        ----------
        row : DataRow
            row to return the dataset names for

        Returns
        -------
        dataset_names : list[str]
            list of dataset names stored in the given row
        """
        dataset_names = [None]  # The source data
        derivs_dir = (
            Path(row.frameset.id) / self._row_relpath(row, dataset_name="").parent
        )
        if derivs_dir.exists():
            dataset_names.extend(
                ("" if d.name == self.EMPTY_DATASET_NAME else d.name)
                for d in derivs_dir.iterdir()
                if d.name != self.FRAMETREE_DIR
            )
        return dataset_names

    def _fileset_fspath(self, entry: DataEntry) -> Path:
        return Path(entry.row.frameset.id) / entry.uri

    def _fields_fspath_and_key(self, entry: DataEntry) -> ty.Tuple[Path, str]:
        relpath, key = entry.uri.split("::")
        fspath = Path(entry.row.frameset.id) / relpath
        return fspath, key

    def _fileset_prov_fspath(self, entry: DataEntry) -> Path:
        return self._fileset_fspath(entry).with_suffix(self.PROV_SUFFIX)

    def _fields_prov_fspath_and_key(self, entry: DataEntry) -> ty.Tuple[Path, str]:
        fields_fspath, key = self._fields_fspath_and_key(entry)
        return fields_fspath.parent / self.FIELDS_PROV_FNAME, key



from __future__ import annotations
from dataclasses import is_dataclass, fields as dataclass_fields
from typing import Sequence
import typing as ty
import functools
import operator
from types import TracebackType, UnionType
from enum import Enum
import builtins
from copy import copy
import json
import re
import inspect
from importlib import import_module
from inspect import isclass, isfunction
from pathlib import PurePath, Path
import logging
import attrs
from fileformats.core import from_mime, to_mime, DataType
import fileformats.field
import pydra.compose.base
import pydra.compose.python
import pydra.compose.workflow
from pydra.utils import get_fields
from pydra.utils.typing import optional_type, is_fileset_or_union
from pydra.engine.workflow import Workflow
from pydra.utils.typing import TypeParser, is_lazy
from pydra.engine.lazy import LazyField
from frametree.core.exceptions import (
    FrameTreeUsageError,
    FrametreeCannotSerializeDynamicDefinitionError,
)
from .packaging import pkg_versions, package_from_module
from .utils import add_exc_note
from frametree.core import PACKAGE_NAME


logger = logging.getLogger("frametree")


FRAMETREE_PIP = "git+ssh://git@github.com/australian-imaging-service/frametree.git"

HASH_CHUNK_SIZE = 2**20  # 1MB in calc. checksums to avoid mem. issues


@attrs.define
class _FallbackContext:
    """Used to specify that class resolution is permitted to fail within this context
    and return just a string (i.e. in build environments where the required modules
    aren't installed)
    """

    permit: bool = False

    def __enter__(self) -> None:
        self.permit = True

    def __exit__(
        self,
        exception_type: ty.Optional[ty.Type[BaseException]],
        exception_value: ty.Optional[BaseException],
        traceback: ty.Optional[TracebackType],
    ) -> None:
        self.permit = False


@attrs.define
class ClassResolver:
    """
    Parameters
    ----------
    base_class : type
        the target class to resolve the string representation to
    allow_none : bool
        whether None is a valid value, if True, None is returned if the string is None
    alternative_types : list[type]
        alternative types that are allowed to be resolved to
    package : str
        the package to resolve the class from if a partial path is given
    resolve_from_module : bool
        whether to resolve the class from the module if only the module name is given
    """

    base_class: ty.Optional[type] = None
    allow_none: bool = False
    alternative_types: ty.List[type] = attrs.field(factory=list)
    package: str = PACKAGE_NAME

    def __call__(self, class_str: str) -> type:
        """
        Resolves a class from a location string in the format "<module-name>:<class-name>"

        Parameters
        ----------
        class_str : str
            Module path and name of class joined by ':', e.g. main_pkg.sub_pkg:MyClass

        Returns
        -------
        type:
            The resolved class
        """
        if class_str is None and self.allow_none:
            return None
        # If only provided with name, attempt to treat it as a submodule/package of the package
        # and there is only one subclass of the base class in the module
        if isinstance(class_str, str) and ":" not in class_str and "/" not in class_str:
            try:
                mod = import_module(self.package + "." + class_str)
            except ModuleNotFoundError:
                pass
            else:
                mod_attrs = [getattr(mod, name) for name in dir(mod)]
                classes = [
                    a
                    for a in mod_attrs
                    if isclass(a) and issubclass(a, self.base_class)
                ]
                if not classes:
                    raise ValueError(
                        f"Did not find a class in module '{class_str}' that is a "
                        f"subclass of {self.base_class}"
                    )
                if len(classes) > 1:
                    superclasses = [
                        c for c in classes if all(issubclass(sc, c) for sc in classes)
                    ]
                    if len(superclasses) == 1:
                        return superclasses[0]
                    raise ValueError(
                        f"Found multiple classes in module '{class_str}' that are "
                        f"subclasses of {self.base_class}: {classes}"
                    )
                return classes[0]
        klass = self.fromstr(class_str, subpkg=True, pkg=self.package)
        base_class = optional_type(self.base_class)
        if (
            inspect.isclass(base_class)
            and inspect.isclass(klass)
            and issubclass(base_class, DataType)
            and not issubclass(klass, DataType)
        ):
            try:
                klass = fileformats.field.Field.from_primitive(klass)
            except TypeError:
                pass
        self._check_type(klass)
        return klass

    @classmethod
    def _get_subpkg(cls, klass: ty.Type[ty.Any]) -> ty.Optional[str]:
        try:
            return klass.SUBPACKAGE
        except AttributeError:
            return None

    @classmethod
    def fromstr(
        cls, class_str: str, subpkg: ty.Optional[str] = None, pkg: str = PACKAGE_NAME
    ) -> ty.Union[ty.Type[ty.Any], ty.Callable]:
        """Resolves a class/function from a string containing its module an its name
        separated by a ':'

        Parameters
        ----------
        class_str : str
            the string representation to resolve to a class or function
        subpkg : str, optional
            the sub-package that the class should belong to within the extension

        Returns
        -------
        type or callable
            the resolved class or function

        Raises
        ------
        ValueError
            raised if the string doesn't contain a ':'
        FrameTreeUsageError
            raised if the class wasn't found in the sub-package
        FrameTreeUsageError
            raised if a sub-package couldn't be found
        """
        if not isinstance(class_str, str):
            return class_str  # Assume that it is already resolved
        if "/" in class_str:  # Assume mime-type/like string
            return from_mime(class_str)
        if (
            "|" in class_str
        ):  # Assume union type; option 3: use functools.reduce with operator.or_
            union_args = tuple(
                cls.fromstr(t.strip(), subpkg=subpkg, pkg=pkg)
                for t in class_str.split("|")
            )
            return functools.reduce(operator.or_, union_args)
        if class_str.startswith("<") and class_str.endswith(">"):
            class_str = class_str[1:-1]
        try:
            module_path, class_name = class_str.split(":")
        except ValueError:
            try:
                return getattr(builtins, class_str)
            except AttributeError:
                module_path = None
                class_name = class_str
        assumed_common = False
        if not module_path:
            module_path = "common"  # default package
            assumed_common = True
        if "." in module_path:
            # Interpret as an absolute path not a relative path from an extension
            module_path = module_path.rstrip(
                "."
            )  # trailing '.' signifies top-level pkg
            subpkg = None
        module = None

        if subpkg:
            full_mod_path = ".".join((pkg, module_path))
            if isinstance(subpkg, str):
                full_mod_path += "." + subpkg
        else:
            full_mod_path = module_path
        try:
            module = import_module(full_mod_path)
        except ModuleNotFoundError:
            if cls.FALLBACK_TO_STR.permit:
                return class_str
            else:
                msg = (
                    f"Did not find module {full_mod_path}' when resolving {class_str} "
                    f"with subpkg={subpkg}.\n"
                )
                if assumed_common:
                    msg += (
                        "NB: No module path was provided, so the default 'common' "
                        "package was assumed. Please check it isn't meant to be a builtin "
                        "class or function instead"
                    )
                raise FrameTreeUsageError(msg)
        try:
            klass = getattr(module, class_name)
        except AttributeError:
            raise FrameTreeUsageError(
                f"Did not find '{class_str}' class/function in module '{module.__name__}'"
            )
        return klass

    @classmethod
    def tostr(cls, klass: ty.Type[ty.Any], strip_prefix: bool = True) -> str:
        """Records the location of a class so it can be loaded later using
        `ClassResolver`, in the format <module-name>:<class-name>

        Parameters
        ----------
        klass : Any
            the class/function to serialise to a string
        strip_prefix : bool
            whether to strip the SUBPACKAGE prefix from the module path when writing
            to file
        """
        if isinstance(klass, str):
            return klass
        if ty.get_origin(klass) is ty.Union or type(klass) is UnionType:
            if ty.get_origin(klass):
                args = ty.get_args(klass)
            else:
                args = klass.__args__
            return " | ".join(
                cls.tostr(t) if t is not type(None) else "None" for t in args
            )
        if not (isclass(klass) or isfunction(klass)):
            klass = type(klass)  # Get the class rather than the object
        if isclass(klass) and issubclass(klass, DataType):
            return to_mime(klass, official=False)

        module_name = get_module_name(klass)
        if module_name == "builtins":
            return klass.__name__
        if strip_prefix and cls._get_subpkg(klass):
            subpkg = cls._get_subpkg(klass)
            if match := re.match(r"frametree\.(\w+)\." + subpkg, module_name):
                module_name = match.group(
                    1
                )  # just use the name of the extension module
            elif "." not in module_name:
                module_name += "."  # To distinguish it from extension module name
        return module_name + ":" + klass.__name__

    def _check_type(self, klass: ty.Type[ty.Any]) -> None:
        if self.FALLBACK_TO_STR.permit and isinstance(klass, str):
            return
        if self.base_class:
            if isfunction(klass):
                if ty.Callable in self.alternative_types:
                    return  # ok
                else:
                    raise ValueError(
                        f"Found callable {klass}, but Callable isn't in alternative_types"
                    )
            if klass in self.alternative_types:
                return  # ok
            # TypeParser handles unions and other exotic base classes Python < 3.10
            if not (
                TypeParser.is_subclass(klass, self.base_class)
                or issubclass(klass, self.base_class)
            ):
                raise ValueError(
                    f"Found {klass}, which is not a subclass of {self.base_class}"
                )

    FALLBACK_TO_STR = _FallbackContext()


def asdict(
    obj: ty.Any,
    omit: ty.Iterable[str] = (),
    required_modules: ty.Optional[ty.Set[str]] = None,
) -> ty.Dict[str, ty.Any]:
    """Serialises an object of a class defined with attrs to a dictionary

    Parameters
    ----------
    obj
        The FrameTree object to asdict. Must be defined using the attrs
        decorator
    omit: Iterable[str]
        the names of attributes to omit from the dictionary
    required_modules: set[str], optional
        modules required to reload the serialised object into memory

    Returns
    -------
    dict
        the serialised object
    """

    def filter(atr: attrs.Attribute[ty.Any], value: ty.Any) -> bool:
        return (
            atr.init and atr.metadata.get("asdict", True) and value is not attrs.NOTHING
        )

    if required_modules is None:
        required_modules = set()
        include_versions = True  # Assume top-level dictionary so need to include
    else:
        include_versions = False

    def serialise_class(klass: ty.Type[ty.Any]) -> str:
        required_modules.add(get_module_name(klass))
        return "<" + ClassResolver.tostr(klass, strip_prefix=False) + ">"

    def value_asdict(value: ty.Any) -> ty.Dict[str, ty.Any]:
        if is_fileset_or_union(value):
            value = to_mime(value, official=False)
        elif isclass(value):
            value = serialise_class(value)
        elif hasattr(value, "asdict"):
            value = value.asdict(required_modules=required_modules)
        elif attrs.has(value):  # is class with attrs
            value_class = serialise_class(type(value))
            value = attrs.asdict(
                value,
                recurse=False,
                filter=filter,
                value_serializer=lambda i, f, v: value_asdict(v),
            )
            _replace_hidden(value)
            value["class"] = value_class
        elif isinstance(value, Enum):
            value = serialise_class(type(value)) + "[" + str(value) + "]"
        elif isinstance(value, PurePath):
            value = "file://" + str(value.resolve())
        elif isinstance(value, (tuple, list, set, frozenset)):
            value = [value_asdict(x) for x in value]
        elif isinstance(value, dict):
            value = {value_asdict(k): value_asdict(v) for k, v in value.items()}
        elif is_dataclass(value):
            value = [
                value_asdict(getattr(value, f.name)) for f in dataclass_fields(value)
            ]
        return value

    dct = attrs.asdict(
        obj,
        recurse=False,
        filter=lambda a, v: filter(a, v) and a.name not in omit,
        value_serializer=lambda i, f, v: value_asdict(v),
    )

    _replace_hidden(dct)

    dct["class"] = serialise_class(type(obj))
    if include_versions:
        dct["pkg_versions"] = pkg_versions(required_modules)

    return dct


def _replace_hidden(dct: ty.Dict[str, ty.Any]) -> None:
    "Replace hidden attributes (those starting with '_') with non-hidden"
    for key in list(dct):
        if key.startswith("_"):
            dct[key[1:]] = dct.pop(key)


def fromdict(dct: ty.Dict[str, ty.Any], **kwargs: ty.Any) -> object:
    """Unserialise an object from a dict created by the `asdict` method

    Parameters
    ----------
    dct : dict
        A dictionary containing a serialsed FrameTree object such as a data store
        or dataset definition
    omit: Iterable[str]
        key names to ignore when unserialising
    **kwargs : dict[str, Any]
        Additional initialisation arguments for the object when it is reinitialised.
        Overrides those stored"""
    # try:
    #     frametree_version = dct["pkg_versions"]["frametree"]
    # except (TypeError, KeyError):
    #     pass
    #     else:
    #         if (packaging.version.parse(frametree_version)
    #               < packaging.version.parse(MIN_SERIAL_VERSION)):
    #             raise FrameTreeVersionError(
    #                 f"Serialised version ('{frametree_version}' is too old to be "
    #                 f"read by this version of frametree ('{__version__}'), the minimum "
    #                 f"version is {MIN_SERIAL_VERSION}")

    def field_filter(klass: ty.Type[ty.Any], field_name: str) -> bool:
        if attrs.has(klass):
            return field_name in (f.name for f in attrs.fields(klass))
        else:
            return field_name != "class"

    def fromdict(
        value: ty.Union[
            ty.Dict[str, ty.Any],
            str,
            ty.Sequence[ty.Dict[str, ty.Any]],
        ],
    ) -> ty.Any:
        resolved_value: ty.Any = value
        if isinstance(value, dict):
            if "class" in value:
                klass = ClassResolver()(value["class"])
                if hasattr(klass, "fromdict"):
                    return klass.fromdict(value)
            resolved_value = {fromdict(k): fromdict(v) for k, v in value.items()}
            if "class" in resolved_value:
                resolved_value = klass(
                    **{
                        k: v
                        for k, v in resolved_value.items()
                        if field_filter(klass, k)
                    }
                )
        elif isinstance(value, str):
            if match := re.match(r"<(.*)>$", value):  # Class location
                resolved_value = ClassResolver()(match.group(1))
            elif match := re.match(
                r"<(.*)>\[(.*)\]$", value
            ):  # Enum or classified format
                resolved_value = ClassResolver()(match.group(1))[match.group(2)]  # type: ignore[index]
            elif match := re.match(r"file://(.*)", value):
                resolved_value = Path(match.group(1))
        elif isinstance(value, Sequence):
            resolved_value = [fromdict(x) for x in value]
        return resolved_value

    klass = ClassResolver()(dct["class"])

    kwargs.update(
        {
            k: fromdict(v)
            for k, v in dct.items()
            if field_filter(klass, k) and k not in kwargs
        }
    )

    return klass(**kwargs)


extract_import_re = re.compile(r"\s*(?:from|import)\s+([\w\.]+)")

NOTHING_STR = "__PIPELINE_INPUT__"


def pydra_asdict(
    obj: pydra.compose.base.Task,
    required_modules: ty.Set[str],
    workflow: ty.Optional[Workflow] = None,
) -> ty.Dict[str, ty.Any]:
    """Converts a Pydra Task/Workflow into a dictionary that can be serialised

    Parameters
    ----------
    obj : pydra.compose.base.Task
        the Pydra object to convert to a dictionary
    required_modules : set[str]
        a set of modules that are required to load the pydra object back
        out from disk and run it
    workflow : pydra.Workflow, optional
        the containing workflow that the object to serialised is part of

    Returns
    -------
    dict
        the dictionary containing the contents of the Pydra object
    """
    dct = {
        "class": "<" + ClassResolver.tostr(obj, strip_prefix=False) + ">",
    }
    if isinstance(obj, Workflow):
        dct["nodes"] = [
            pydra_asdict(n, required_modules=required_modules, workflow=obj)
            for n in obj.nodes
        ]
        dct["outputs"] = outputs = {}
        for outpt_name, lf in obj._connections:
            outputs[outpt_name] = {"task": lf.name, "field": lf.field}
    else:
        if isinstance(obj, (pydra.compose.python.Task, pydra.compose.workflow.Task)):
            klass = type(obj)
            func = (
                obj.function
                if isinstance(obj, pydra.compose.python.Task)
                else obj.constructor
            )
            if klass.__module__ == "types":
                mod_name = func.__module__
            else:
                mod_name = klass.__module__
            dct["class"] = "<" + mod_name + ":" + klass.__name__ + ">"
            required_modules.add(mod_name)
            # inspect source for any import lines (should be present in function
            # not module)
            for line in inspect.getsourcelines(func)[0]:
                if match := extract_import_re.match(line):
                    required_modules.add(match.group(1))
            # TODO: check source for references to external modules that aren't
            #       imported within function
        else:
            pkg = package_from_module(type(obj).__module__)
            dct["package"] = pkg.key
            dct["version"] = pkg.version
        if hasattr(obj, "container"):
            dct["container"] = {"type": obj.container, "image": obj.image}
    dct["inputs"] = inputs = {}
    for inpt in get_fields(obj):
        if inpt.name.startswith("_"):
            continue
        inpt_value = getattr(obj, inpt.name)
        if (
            obj._task_type() == "python"
            and inpt.name == "function"
            and inpt_value is inpt.default
        ):
            # Don't include the function in the serialised object
            continue
        if (
            obj._task_type() == "workflow"
            and inpt.name == "constructor"
            and inpt_value is inpt.default
        ):
            # Don't include the constructor in the serialised object
            continue
        if is_lazy(inpt_value):
            inputs[inpt.name] = {"field": inpt_value.field}
            # If the lazy field comes from the workflow lazy in, we omit
            # the "task" item
            if workflow is None or inpt_value.name != workflow.name:
                inputs[inpt.name]["task"] = inpt_value.name
        elif inpt_value == attrs.NOTHING:
            inputs[inpt.name] = NOTHING_STR
        else:
            inputs[inpt.name] = inpt_value
    return dct


def lazy_field_fromdict(dct: ty.Dict[ty.Any, ty.Any], workflow: Workflow) -> LazyField:
    """Unserialises a LazyField object from a dictionary"""
    if "task" in dct:
        inpt_task = getattr(workflow, dct["task"])
        lf = getattr(inpt_task.lzout, dct["field"])
    else:
        lf = getattr(workflow.lzin, dct["field"])
    return lf


def pydra_fromdict(
    dct: ty.Dict[ty.Any, ty.Any],
    workflow: ty.Optional[Workflow] = None,
    **kwargs: ty.Any,
) -> pydra.compose.base.Task:
    """Recreates a Pydra Task/Workflow from a dictionary object created by
    `pydra_asdict`

    Parameters
    ----------
    dct : dict
        dictionary representations of the object to recreate
    name : str
        name to give the object
    workflow : pydra.Workflow, optional
        the containing workflow that the object to recreate is connected to
    **kwargs
        additional keyword arguments passed to the pydra Object init method

    Returns
    -------
    pydra.compose.base.Task
        the recreated Pydra object
    """
    klass = ClassResolver()(dct["class"])
    # Resolve lazy-field references to workflow fields
    inputs = {}
    for inpt_name, inpt_val in dct["inputs"].items():
        if inpt_val == NOTHING_STR:
            continue
        # Check for 'field' key in a dictionary val and convert to a
        # LazyField object
        if isinstance(inpt_val, dict) and "field" in inpt_val:
            inpt_val = lazy_field_fromdict(inpt_val, workflow=workflow)
        inputs[inpt_name] = inpt_val
    kwargs.update((k, v) for k, v in inputs.items() if k not in kwargs)
    if klass is Workflow:
        obj = Workflow(name=dct["name"], input_spec=list(dct["inputs"]), **kwargs)
        for node_dict in dct["nodes"]:
            obj.add(pydra_fromdict(node_dict, workflow=obj))
        obj.set_output(
            [
                (n, lazy_field_fromdict(f, workflow=obj))
                for n, f in dct["outputs"].items()
            ]
        )
    else:
        obj = klass(**kwargs)
    return obj


@attrs.define
class ObjectConverter:

    klass: type
    allow_none: bool = False
    default_if_none: ty.Any = None
    accept_metadata: bool = False
    package: str = PACKAGE_NAME

    def __call__(self, value: ty.Any) -> ty.Any:
        return self._create_object(value)

    def _create_object(self, value: ty.Any, **kwargs: ty.Any) -> ty.Any:
        if value is None:
            if kwargs:
                value = {}
            elif self.allow_none:
                if callable(self.default_if_none):
                    default = self.default_if_none()
                else:
                    default = self.default_if_none
                return default
            else:
                raise ValueError(
                    f"None values not accepted in automatic conversion to {self.klass}"
                )
        if isinstance(value, dict):
            if self.accept_metadata:
                klass_attrs = set(attrs.fields_dict(self.klass))
                value_kwargs = {k: v for k, v in value.items() if k in klass_attrs}
                value_kwargs["metadata"] = {
                    k: v for k, v in value.items() if k not in klass_attrs
                }
            else:
                value_kwargs = value
            value_kwargs.update(kwargs)
            try:
                obj = self.klass(**value_kwargs)
            except TypeError as e:
                field_names = [f.name for f in attrs.fields(self.klass)]
                msg = f"when creating {self.klass} from {value_kwargs}, expected {field_names}"
                add_exc_note(e, msg)
                raise
        elif isinstance(value, (list, tuple)):
            obj = self.klass(*value, **kwargs)
        elif isinstance(value, self.klass):
            obj = copy(value)
            for k, v in kwargs.items():
                setattr(obj, k, v)
        elif isinstance(value, (str, int, float, bool)):
            # If there are kwargs that are in the first N positions of the
            # argument list, add them in as positional arguments first and then
            # append the value to the end of the args list
            args = []
            kgs = copy(kwargs)
            for field_name in attrs.fields_dict(self.klass):
                try:
                    args.append(kgs.pop(field_name))
                except KeyError:
                    break
            args.append(value)
            obj = self.klass(*args, **kgs)
        else:
            raise ValueError(f"Cannot convert {value} into {self.klass}")
        return obj


@attrs.define
class ObjectListConverter(ObjectConverter):
    def __call__(self, value: ty.Any) -> ty.List[ty.Any]:
        converted: ty.List[ty.Any] = []
        if value is None:
            if self.allow_none:
                return converted
            else:
                raise ValueError("Value cannot be None")
        if isinstance(value, dict):
            for name, item in value.items():
                converted.append(self._create_object(item, name=name))
        else:
            for item in value:
                converted.append(self._create_object(item))
        return converted

    @classmethod
    def asdict(cls, objs: ty.List[ty.Any], **kwargs: ty.Any) -> ty.Dict[str, ty.Any]:
        dct = {}
        for obj in objs:
            obj_dict = attrs.asdict(obj, **kwargs)
            dct[obj_dict.pop("name")] = obj_dict
        return dct

    @classmethod
    def aslist(cls, objs: ty.List[ty.Any], **kwargs: ty.Any) -> ty.List[ty.Any]:
        return [attrs.asdict(obj, **kwargs) for obj in objs]


def parse_value(value: ty.Any) -> ty.Any:
    """Parses values from string representations"""
    try:
        value = json.loads(
            value
        )  # FIXME: Is this value replace really necessary, need to investigate where it is used again
    except (TypeError, json.decoder.JSONDecodeError):
        pass
    return value


def get_module_name(klass: type) -> str:
    """Gets the module in which the klass was defined, taking into account dynamically
    created Pydra Task classes"""
    if klass.__module__ == "types":
        try:
            executor_name = klass._executor_name
        except AttributeError:
            pass
        else:
            executor = get_fields(klass)[executor_name].default
            try:
                module_name = executor.__module__
            except AttributeError:
                pass
            else:
                if module_name != "types":
                    module = import_module(module_name)
                    if getattr(module, klass.__name__) is klass:
                        return module_name
        raise FrametreeCannotSerializeDynamicDefinitionError(
            f"Cannot serialise {klass} as it is a dynamically created class"
        )
    return klass.__module__



from __future__ import annotations
from pathlib import Path
import re
from abc import abstractmethod
import typing as ty
import errno
import logging
import json
import attrs
import yaml
from fasteners import InterProcessLock
from fileformats.core import DataType, FileSet, Field
from frametree.core.exceptions import (
    FrameTreeMissingDataException,
    FrameTreeUsageError,
    DatatypeUnsupportedByStoreError,
)
from frametree.core.utils import get_home_dir, append_suffix
from ..row import DataRow
from ..entry import DataEntry
from .base import Store


logger = logging.getLogger("frametree")


# Matches directory names used for summary rows with dunder beginning and
# end (e.g. '__visit_01__') and hidden directories (i.e. starting with '.' or
# '~')
special_dir_re = re.compile(r"(__.*__$|\..*|~.*)")

DT = ty.TypeVar("DT", bound=DataType)


@attrs.define
class LocalStore(Store):
    """
    A Repository class for data stored hierarchically within sub-directories
    of a file-system directory. The depth and which layer in the data tree
    the sub-directories correspond to is defined by the `hierarchy` argument.

    Parameters
    ----------
    base_dir : str
        Path to the base directory of the "store", i.e. datasets are
        arranged by name as sub-directories of the base dir.

    """

    LOCK_SUFFIX = ".lock"
    FRAMETREE_DIR = "__frametree__"
    SITE_LICENSES_DIR = "site-licenses"

    name: str

    ##############################
    # Inherited abstract methods #
    ##############################

    # populate_tree

    # populate_row

    # put_provenance

    # get_provenance

    # create_data_tree

    ####################
    # Abstract methods #
    ####################

    @abstractmethod
    def get_field(self, entry: DataEntry, datatype: type) -> Field[ty.Any, ty.Any]:
        """Retrieves a field from a data entry

        Parameters
        ----------
        entry : DataEntry
            the entry to retrieve the file-set from
        datatype : type
            the type of the field from

        Returns
        -------
        Field
            the retrieved field
        """

    @abstractmethod
    def put_fileset(self, fileset: FileSet, entry: DataEntry) -> FileSet:
        """Stores a file-set into a data entry

        Parameters
        ----------
        fileset : FileSet
            the file-set to store
        entry : DataEntry
            the entry to store the file-set in

        Returns
        -------
        FileSet
            the file-set within the store
        """

    @abstractmethod
    def get_fileset(self, entry: DataEntry, datatype: type) -> FileSet:
        """Retrieves a file-set from a data entry

        Parameters
        ----------
        entry : DataEntry
            the entry to retrieve the file-set from
        datatype : type
            the type of the file-set

        Returns
        -------
        FileSet
            the retrieved file-set
        """

    @abstractmethod
    def put_field(self, field: Field[ty.Any, ty.Any], entry: DataEntry) -> None:
        """Stores a field into a data entry

        Parameters
        ----------
        field : Field
            the field to store
        entry : DataEntry
            the entry to store the field in
        """

    @abstractmethod
    def fileset_uri(self, path: str, datatype: type, row: DataRow) -> str:
        """Returns the "uri" (e.g. file-system path relative to root dir) of a file-set
        entry at the given path relative to the given row

        Parameters
        ----------
        path : str
            path to the entry relative to the row
        datatype : type
            the datatype of the entry
        row : DataRow
            the row of the entry

        Returns
        -------
        uri : str
            the "uri" to the file-set entry relative to the data store
        """

    @abstractmethod
    def field_uri(self, path: str, datatype: type, row: DataRow) -> str:
        """Returns the "uri" (e.g. file-system path relative to root dir) of a field
        entry at the given path relative to the given row

        Parameters
        ----------
        path : str
            path to the entry relative to the row
        datatype : type
            the datatype of the entry
        row : DataRow
            the row of the entry

        Returns
        -------
        uri : str
            the "uri" to the field entry relative to the data store
        """

    @abstractmethod
    def get_fileset_provenance(
        self, entry: DataEntry
    ) -> ty.Union[ty.Dict[str, ty.Any], None]:
        """Retrieves provenance associated with a file-set data entry

        Parameters
        ----------
        entry : DataEntry
            the entry of the file-set to retrieve the provenance for

        Returns
        -------
        ty.Dict[str, ty.Any] or None
            the retrieved provenance
        """

    @abstractmethod
    def put_fileset_provenance(
        self, provenance: ty.Dict[str, ty.Any], entry: DataEntry
    ) -> None:
        """Puts provenance associated with a file-set data entry into the store

        Parameters
        ----------
        provenance : dict[str, ty.Any]
            the provenance to store
        entry : DataEntry
            the entry to associate the proveance with
        """

    @abstractmethod
    def get_field_provenance(
        self, entry: DataEntry
    ) -> ty.Union[ty.Dict[str, ty.Any], None]:
        """Retrieves provenance associated with a field data entry

        Parameters
        ----------
        entry : DataEntry
            the entry of the field to retrieve the provenance for

        Returns
        -------
        ty.Dict[str, ty.Any] or None
            the retrieved provenance
        """

    @abstractmethod
    def put_field_provenance(
        self, provenance: ty.Dict[str, ty.Any], entry: DataEntry
    ) -> None:
        """Puts provenance associated with a field data entry into the store

        Parameters
        ----------
        provenance : dict[str, ty.Any]
            the provenance to store
        entry : DataEntry
            the entry to associate the proveance with
        """

    ##################################
    # Abstractmethod implementations #
    ##################################

    def connect(self) -> None:
        return None

    def disconnect(self, connection: ty.Any) -> None:
        pass

    def create_entry(self, path: str, datatype: type, row: DataRow) -> DataEntry:
        if issubclass(datatype, FileSet):
            uri = self.fileset_uri(path, datatype, row)
        else:
            uri = self.field_uri(path, datatype, row)
        return row.add_entry(path=path, datatype=datatype, uri=uri)

    def save_frameset_definition(
        self, dataset_id: str, definition: ty.Dict[str, ty.Any], name: str
    ) -> None:
        definition_path = self.definition_save_path(dataset_id, name)
        definition_path.parent.mkdir(exist_ok=True, parents=True)
        with open(definition_path, "w") as f:
            yaml.dump(definition, f)

    def load_frameset_definition(
        self, dataset_id: str, name: str
    ) -> ty.Dict[str, ty.Any]:
        fspath = self.definition_save_path(dataset_id, name)
        if fspath.exists():
            with open(fspath) as f:
                definition = yaml.load(f, Loader=yaml.Loader)
        else:
            definition = None
        return definition

    def get(self, entry: DataEntry, datatype: ty.Type[DT]) -> DT:
        if entry.datatype.is_fileset:
            item = self.get_fileset(entry, datatype)
        elif entry.datatype.is_field:
            item = self.get_field(entry, datatype)
        else:
            raise RuntimeError(
                f"Don't know how to retrieve {entry.datatype} data from {type(self)} stores"
            )
        return item

    def put(self, item: DT, entry: DataEntry) -> DT:
        if entry.datatype.is_fileset:
            cpy = self.put_fileset(item, entry)
        elif entry.datatype.is_field:
            cpy = self.put_field(item, entry)
        else:
            raise RuntimeError(
                f"Don't know how to store {entry.datatype} data in {type(self)} stores"
            )
        return cpy

    def get_provenance(self, entry: DataEntry) -> ty.Dict[str, ty.Any]:
        if entry.datatype.is_fileset:
            provenance = self.get_fileset_provenance(entry)
        elif entry.datatype.is_field:
            provenance = self.get_field_provenance(entry)
        else:
            raise DatatypeUnsupportedByStoreError(entry.datatype, self)
        return provenance

    def put_provenance(
        self, provenance: ty.Dict[str, ty.Any], entry: DataEntry
    ) -> None:
        if entry.datatype.is_fileset:
            self.put_fileset_provenance(provenance, entry)
        elif entry.datatype.is_field:
            self.put_field_provenance(provenance, entry)
        else:
            raise DatatypeUnsupportedByStoreError(entry.datatype, self)

    def root_dir(self, row: DataRow) -> Path:
        return Path(row.frameset.id)

    def site_licenses_dataset(self, **kwargs: ty.Any):
        """Provide a place to store hold site-wide licenses"""
        dataset_root = get_home_dir() / self.SITE_LICENSES_DIR
        if not dataset_root.exists():
            dataset_root.mkdir(parents=True)
        try:
            dataset = self.load_frameset(dataset_root)
        except KeyError:
            from frametree.common import Samples

            dataset = self.define_frameset(dataset_root, axes=Samples)
        return dataset

    ###################
    # Other overrides #
    ###################

    def define_frameset(self, id: str, *args: ty.Any, **kwargs: ty.Any):
        if not Path(id).exists():
            raise FrameTreeUsageError(f"Path to dataset root '{id}'' does not exist")
        return super().define_frameset(id, *args, **kwargs)

    ##################
    # Helper methods #
    ##################

    def update_json(self, fpath: Path, key: str, value: ty.Any) -> None:
        """Updates a JSON file in a multi-process safe way"""
        # Open fields JSON, locking to prevent other processes
        # reading or writing
        with InterProcessLock(append_suffix(fpath, self.LOCK_SUFFIX), logger=logger):
            try:
                with open(fpath) as f:
                    dct = json.load(f)
            except IOError as e:
                if e.errno == errno.ENOENT:
                    dct = {}
                else:
                    raise
            dct[key] = value
            with open(fpath, "w") as f:
                json.dump(dct, f, indent=4)

    def read_from_json(self, fpath: Path, key: str) -> ty.Any:
        """
        Load fields JSON, locking to prevent read/write conflicts
        Would be better if only checked if locked to allow
        concurrent reads but not possible with multi-process
        locks (in my understanding at least).
        """
        try:
            with InterProcessLock(
                append_suffix(fpath, self.LOCK_SUFFIX), logger=logger
            ), open(fpath, "r") as f:
                dct = json.load(f)
            return dct[key]
        except (KeyError, IOError) as e:
            try:
                # Check to see if the IOError wasn't just because of a
                # missing file
                if e.errno != errno.ENOENT:
                    raise
            except AttributeError:
                pass
            raise FrameTreeMissingDataException(
                "{} does not exist in the local store {}".format(key, self)
            )

    def definition_save_path(self, dataset_id: str, name: str) -> Path:
        if not name:
            name = "_"
        return Path(dataset_id) / self.FRAMETREE_DIR / name / "definition.yaml"



import re
import pytest
from frametree.core.packaging import package_from_module
from frametree.core.utils import path2varname, varname2path


def test_package_from_module():
    assert package_from_module("frametree.core").key == "frametree"
    assert package_from_module("pydra.engine").key == "pydra"


PATHS_TO_TEST = [
    "dwi/dir-LR_dwi",
    "func/task-rest_bold",
    "with spaces and___ underscores",
    "__a.very$illy*ath~",
    "anat/T1w",
    "anat___l___T1w",
    "_u__u_",
]


@pytest.mark.parametrize("path", PATHS_TO_TEST)
def test_path2varname(path: str):
    varname = path2varname(path)
    assert re.match(r"^\w+$", varname)
    assert varname2path(varname) == path


@pytest.mark.parametrize("path", PATHS_TO_TEST)
def test_triple_path2varname(path: str):
    assert (
        varname2path(
            varname2path(varname2path(path2varname(path2varname(path2varname(path)))))
        )
        == path
    )



from frametree.core.axes import Axes


class Clinical(Axes):
    """
    An enum used to specify the hierarchy of data trees and the "frequencies" of items
    within dataset typical of medimage research, i.e. subjects split into groups scanned
    at different visits (in longitudinal studies).
    """

    # Root row of the dataset
    constant = 0b000  # constant across the dataset

    # Axes of the data space
    member = 0b001  # subjects relative to their group membership, i.e.
    # matched pairs of test and control subjects should share
    # the same member IDs.
    group = 0b010  # subject groups (e.g. test & control)
    visit = 0b100  # visits in longitudinal studies

    # Combinations
    session = 0b111  # a single session (i.e. a single visit of a subject)
    subject = 0b011  # uniquely identified subject within in the dataset.
    # As opposed to 'member', which specifies a subject in
    # relation to its group (i.e. one subject for each member
    # in each group). For datasets with only one study group,
    # then subject and member are equivalent
    groupedvisit = 0b110  # data from separate groups at separate visits
    matchedvisit = 0b101  # matched members (e.g. test & control) across
    # all groups and visits


class Samples(Axes):
    """
    The most basic data space within only one dimension
    """

    # Root row of the dataset
    constant = 0b0  # constant across the dataset

    # Axes of the data space
    sample = 0b1



import shutil
from pathlib import Path
import typing as ty
import attrs
from pydra.compose import python, workflow
import fileformats.core
from fileformats.text import TextFile
import fileformats.text
from frametree.core.row import DataRow


@python.define
def Add(a: float, b: float) -> float:
    return a + b


@python.define
def PathManip(dpath: Path, fname: str) -> tuple[str, str]:
    """
    Returns
    -------
    path : str
        extracted path
    suffix : str
        the extracted suffix
    """
    path = dpath / fname
    return str(path), path.suffix


@attrs.define(auto_attribs=True)
class A:
    x: int
    y: int


@attrs.define(auto_attribs=True)
class B:
    u: float
    v: float


@attrs.define(auto_attribs=True)
class C:
    z: float


@python.define
def AttrsFunc(a: A, b: B) -> C:
    """A function that takes two attrs classes and returns a third

    Parameters
    ----------
    a : A
        an attrs class
    b : B
        another attrs class

    Returns
    -------
    c : C
        a third attrs class
    """
    return C(z=a.x * b.u + a.y * b.v)


@python.define(outputs=["out_file"])
def Concatenate(
    in_file1: TextFile,
    in_file2: TextFile,
    duplicates: int = 1,
) -> TextFile:
    """Concatenates the contents of two files and writes them to a third

    Parameters
    ----------
    in_file1 : Path
        A text file
    in_file2 : Path
        Another text file

    Returns
    -------
    out_file: Path
        A text file made by concatenating the two inputs
    """
    out_file = Path("out_file.txt").absolute()
    contents = []
    for _ in range(duplicates):
        for fname in (in_file1, in_file2):
            with open(fname) as f:
                contents.append(f.read())
    with open(out_file, "w") as f:
        f.write("\n".join(contents))
    return out_file


@python.define(outputs=["out_file"])
def Reverse(in_file: TextFile) -> TextFile:
    """Reverses the contents of a file and outputs it to another file

    Parameters
    ----------
    in_file : TextFile
        A text file

    Returns
    -------
    out_file: Path
        A text file with reversed contents to the original
    """
    out_file = Path("out_file.txt").absolute()
    with open(in_file) as f:
        contents = f.read()
    with open(out_file, "w") as f:
        f.write(contents[::-1])
    return out_file


@workflow.define(outputs=["out_file"])
def ConcatenateReverse(
    in_file1: TextFile, in_file2: TextFile, duplicates: int
) -> TextFile:
    """A simple workflow that has the same signature as concatenate, but
    concatenates reversed contents of the input files instead

    Parameters
    ----------
    name : str
        name of the workflow to be created
    **kwargs
        keyword arguments passed through to the workflow init, can be any of
        the workflow's input spec, i.e. ['in_file1', 'in_file2', 'duplicates']

    Returns
    -------
    out_file: TextFile
        The file made by concatenating the reversed contents of the two inputs
    """

    reverse1 = workflow.add(Reverse(in_file=in_file1), name="reverse1")

    reverse2 = workflow.add(Reverse(in_file=in_file2), name="reverse2")

    concatenate = workflow.add(
        Concatenate(
            in_file1=reverse1.out_file,
            in_file2=reverse2.out_file,
            duplicates=duplicates,
        )
    )

    return concatenate.out_file


@python.define
def Plus10ToFilenumbers(filenumber_row: DataRow) -> None:
    """Alters the item paths within the data row, by converting them to
    an int and adding 10. Used in the test_run_pipeline_on_row_cli test.

    Parameters
    ----------
    row : DataRow
        the data row to modify
    """
    for entry in filenumber_row.entries:
        item = fileformats.text.TextFile(ty.cast(fileformats.core.FileSet, entry.item))
        new_item_stem = str(int(item.stem) + 10)
        shutil.move(item.fspath, item.fspath.parent / (new_item_stem + item.actual_ext))


@python.define
def Identity(in_: ty.Any) -> ty.Any:
    return in_


@python.define
def MultiplyContents(
    in_file: TextFile,
    multiplier: ty.Union[int, float],
    out_file: ty.Optional[Path] = None,
    dtype: type = float,
) -> TextFile:
    """Multiplies the contents of the file, assuming that it contains numeric
    values on separate lines

    Parameters
    ----------
    in_file : Path
        path to input file to multiply the contents of
    multiplier : int or float
        the multiplier to apply to the file values
    out_file : Path
        the path to write the output file to
    dtype : type
        the type to cast the file contents to"""

    if out_file is None:
        out_file = Path("out_file.txt").absolute()

    with open(in_file) as f:
        contents = f.read()

    multiplied = []
    for line in contents.splitlines():
        multiplied.append(str(dtype(line.strip()) * multiplier))

    with open(out_file, "w") as f:
        f.write("\n".join(multiplied))

    return TextFile(out_file)


@python.define
def ContentsAreNumeric(in_file: TextFile) -> bool:
    """Checks the contents of a file to see whether each line can be cast to a numeric
    value

    Parameters
    ----------
    in_file : Path
        the path to a text file

    Returns
    -------
    bool
        if all the lines are numeric return True
    """
    with open(in_file) as f:
        contents = f.read()
    try:
        float(contents.strip())
    except ValueError:
        return False
    return True


@python.define
def CheckLicense(
    expected_license_path: TextFile,
    expected_license_contents: TextFile,
) -> TextFile:
    """Checks the `expected_license_path` to see if there is a file with the same contents
    as that of `expected_license_contents`

    Parameters
    ----------
    expected_license_path : File
        path to the expected license file
    expected_license_contents : File
        path containing the contents expected in the expected license file

    Returns
    -------
    File
        passes through the expected license file so the task can be connected back to the
        dataset
    """
    with open(expected_license_contents) as f:
        expected_contents = f.read()
    with open(expected_license_path) as f:
        actual_contents = f.read()
    if expected_contents != actual_contents:
        raise Exception(
            f'License contents "{actual_contents}" did not match '
            f'expected "{expected_contents}"'
        )
    return expected_license_contents


TEST_TASKS = {
    "add": (Add, {"a": 4, "b": 5}, {"out": 9}),
    "path_manip": (
        PathManip,
        {"dpath": Path("/home/foo/Desktop"), "fname": "bar.txt"},
        {"path": "/home/foo/Desktop/bar.txt", "suffix": ".txt"},
    ),
    "attrs_func": (
        AttrsFunc,
        {"a": A(x=2, y=4), "b": B(u=2.5, v=1.25)},
        {"c": C(z=10)},
    ),
}

BASIC_TASKS = ["Add", "PathManip", "AttrsFunc"]

FILE_TASKS = ["Concatenate"]
